#include <stdio.h>
#include <ctype.h>

	/************************************************
	*						*
	*		-- Move --			*
	*						*
	*	Moves disk files around from		*
	*	       here to there.			*
	*						*
	*	T. Jennings 6 Nov. 82			*
	*	  created 6 Nov. 82			*
	*						*
	*						*
	************************************************/

/*	Performs the same function as COPY, but uses the Archive
bit as an argument in transfers.

	MOVE <source_path> [-a ] <dest_path>

	Where <source_path> and <dest_path> are full UNIX like path
names. Wildcards may be used. */

char *sbrk();
char *getmem();

int archive;			/* true if archive option */
int debug;			/* true if debug option, */
int error;
int filecount;

long file_size;
unsigned memsize;		/* allocated memory size, */
char *membuf;			/* ptr to memory block, */
unsigned count;			/* file byte count, */
long size;			/* file size, bytes */

char sourcearg[80],destarg[80];
char source_dir[80],source_name[80];
char dest_dir[80],dest_name[80];

main(argc,argv)
int argc;
char *argv[];
{
int i;
char c;

	if (argc < 2 ) {
		printf("MOVE: Missing arguments\n");
		exit(1);
	}

	debug= 0; 			/* not debugging, */
	archive= 0;
	filecount= 0;

	strcpy(sourcearg,argv[1]);	/* get source file name, */
	strcpy(destarg,"");		/* incase none spec'd */

	for (i= 2; i < argc; i++) {
		if (*argv[i] == '-') {
			++argv[i]; c= *argv[i];
			switch (tolower(c)) {
				case 'a':
					archive= 1;
					break;
				case 'd':
					debug= 1;
					break;
				default:
					printf("MOVE: Unknown option\n");
					break;
			}
		} else 
			 strcpy(destarg,argv[i]);
	}


	allmem();			/* find largest area to allocate */
	memsize= sizmem()-2048;		/* leave room for level 1 i/o */
	membuf = getmem(memsize);
	if (memsize < 1024) {
		cprintf("**** Fatal error: not enough memory ****\007\n\r");
		exit(1);
	}

/* Do a preliminary checkout of the filenames. Make sure no directories
in the paths are wild, etc and set all defaults. */


	makelower(sourcearg); makelower(destarg);
	sep_dir(sourcearg,source_dir,source_name);
	sep_dir(destarg,dest_dir,dest_name);

	if (strlen(source_name) == 0)		/* if blank source name, */
		strcpy(source_name,"*.*");	/* default to all, */

	if (strlen(dest_name) == 0)		/* if destination name blank, */
		strcat(dest_dir,"\\");		/* add slash for append */

/* Now copy all the matching files. */

	strcpy(sourcearg,source_dir); strcat(sourcearg,source_name);
	i= 0; error= 0;
	while (! error && _xfind(sourcearg,source_name,&file_size,0,i)) {
		makelower(source_name);
		error= copy();			/* Copy the files, */
		++i;				/* next file, count 'em, */
	}
	printf("\nCopied %d file(s)\n",filecount);
	exit(0);
}
/* Perform the file copy, with the names specified. The names are in 
two parts, and the actual disk names must be created from these, 
handling wildcards appropriately. 

If the archive option is set, copy the file only if that attribute is
set, and clear it on the source file.  */

copy()
{
char source[80];				/* source pathname */
char dest[80];					/* destination pathname */
int source_handle;
int dest_handle;
int attrib;
char timedate[8];

	strcpy(source,source_dir);		/* assemble the names, */
	strcat(source,source_name);

	strcpy(dest,dest_dir);
	strcat(dest,dest_name);
	if (strlen(dest_name) == 0)
		strcat(dest,source_name);

	if ( !archive || ((attrib= _chmod(source,0,0)) & 0x20)) {
		printf("%-20s%14s %10lu bytes\n",source_dir,source_name,file_size);
		if (strcmp(source,dest) == 0) {		/* ignore identical names */
			printf("\tIdentical source/destination: Ignored\n");
			return(0);
		}
		source_handle= _xopen(source,0);
		if (source_handle == -1) {
			printf("\tIllegal filename: ignored\n",source);
			return(0);
		}
		dest_handle= _xcreate(dest,1);
		if (dest_handle == -1) {
			printf("\tCan't create destination file\n",dest);
			return(1);
		}

		do {	count= _xread(source_handle,membuf,memsize);
			if (_xwrite(dest_handle,membuf,count) != count) {
				printf("\tWrite error: aborted\n");
				return(1);
			}
		} while (count == memsize);
		_ftime(0,source_handle,timedate);	/* transfer the */
		_ftime(1,dest_handle,timedate);		/* creation time */
		_xclose(source_handle);
		_xclose(dest_handle);
		if (archive) {
			attrib &= ~0x20;
			_chmod(source,attrib,1);
		}
		++filecount;
	}
	return(0);
}
/* Seperate the filename from a path spec, return is seperate strings. */

sep_dir(source,dir,name)
char *source;
char *dir;
char *name;
{
int i;
char *ptr;

	strcpy(dir,source);			/* init dir name string, */
	strcpy(name,"");			/* and file name, */
	i= _chmod(dir,0,0);			/* if its a directory, */
	if ((i != -1) && (i & 0x10))
		return;				/* were done */

	i= 0; 
	ptr= dir;				/* work from dir name, */
	while (*ptr)				/* find end of string, */
		++ptr;
	i= strlen(dir);				/* get length, */
	while (i--) {				/* look backwards for a */
		--ptr;
		if ((*ptr == '\\') || (*ptr == '/') || (*ptr == ':')) {
			++ptr;
			break;
		}
	}
	strcpy(name,ptr);			/* copy name part */
	*ptr =0x00;				/* mark end of dir name */
	return;
}
/* Convert a string to lower case. */

makelower(string)
char *string;
{
char c;
	do {	c= tolower(*string);
		*string++ =c;
	} while (c);
	return;
}
/* Return the total number of ? or * characters in the string. */

wild(string)
char *string;
{
int ret;
char c;
	ret= 0;
	do {	c= *string++;
		if (( c == '?') || (c == '*') )
			++ret;
	} while (c);
	return(ret);
}

